/*
 * $Id: challenge.c,v 1.2 2002/10/15 08:41:09 skyper Exp $
 */

#include "common.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include "thcirc.h"
#include "user.h"
#include "system.h"
#include "crypto_func.h"
#include "irc.h"
#include "b64.h"
#include "sanitize.h"
#include "challenge.h"


extern struct _opt opt;
extern char g_buf[4096];

static int AB_sign(USER *u, char *buf, size_t *len);
static int AB_verify(USER *u, char *msg, char *end);

/*
 */
int
A_send(char *nick)
{
	USER *u;
	struct _THCch1 ch1;

	if ( (u = USER_lookup(nick)) == NULL)
	{
		if ( (u = USER_add(nick)) == NULL)
		{
			THCerr("Failed to add user (memory?).\n");
			return -1;
		}
	}
	USER_init(u);
	u->flags |= USER_FL_ALICE;  /* We are Alice */

	if (opt.privKey == NULL)
	{
		THCnfo("Use /loadkey <password> first.\n");
		return -1;
	}

	if (get_random(u->usm->A_noince, sizeof u->usm->A_noince) == NULL)
	{
		THCerr("Unable to get random data: %s\n", strerror(errno));
		return -1;
	}

	u->usm->dh = DH_new();
	if (!u->usm->dh)
	{
		THCerr("%s(): Out of Memory\n", __func__);
		return -1;
	}

	u->usm->dh->p = BN_dup(opt.dhparam->p);
	u->usm->dh->g = BN_dup(opt.dhparam->g);
	if (!DH_generate_key(u->usm->dh))
	{
		THCerr("%s(): DH_generate_key failed\n", __func__);
		return -1;
	}
	if ((BN_num_bytes(u->usm->dh->p) > sizeof ch1.A_dh_p) || (BN_num_bytes(u->usm->dh->pub_key) > sizeof ch1.A_dh_pub))
	{
		THCerr("%s(): Size mismatch\n", __func__);
		return -1;
	}

	memcpy(ch1.A_noince, u->usm->A_noince, sizeof ch1.A_noince);
	BN_bn2bin(u->usm->dh->p, ch1.A_dh_p);
	BN_bn2bin(u->usm->dh->pub_key, ch1.A_dh_pub);

	b64_enc(g_buf, sizeof g_buf, (char *)&ch1, sizeof ch1);
	irc(nick, "K", "%s", g_buf);

	u->state = STATE_PUBSENT;
	return 0;
}


/*
 */
int
B_answer(char *nick, char *msg)
{
	USER *u;
	struct _THCch1 *ch1;
	struct _THCch2 ch2;
	BIGNUM *pub;
	
	if ( (u = USER_lookup(nick)) == NULL)
	{
		if ( (u = USER_add(nick)) == NULL)
		{
			THCerr("Could not add user (list full).\n");
			return -1;
		}
	}
	USER_init(u);

	if (b64_dec(g_buf, sizeof g_buf, msg, strlen(msg)) != sizeof *ch1)
	{
		THCerr("%s(): MSG size mismatch\n", __func__);
		return -1;
	}
	ch1 = (struct _THCch1 *)g_buf;

	/* Save Alice noince and get our own onw */
	memcpy(u->usm->A_noince, ch1->A_noince, sizeof u->usm->A_noince);
	if (get_random(u->usm->B_noince, sizeof u->usm->B_noince) == NULL)
	{
		THCerr("Unable to get random data: %s\n", strerror(errno));
		return -1;
	}

	u->usm->dh = DH_new();
	if (!u->usm->dh)
	{
		THCerr("%s(): Out of memory.\n", __func__);
		return -1;
	}

	u->usm->dh->p = BN_bin2bn(ch1->A_dh_p, sizeof ch1->A_dh_p, NULL);
	u->usm->dh->g = BN_new();
	BN_hex2bn(&u->usm->dh->g, "05");
	if (!DH_generate_key(u->usm->dh))
	{
		THCerr("DH_generate_key failed\n");
		return -1;
	}

	pub = BN_bin2bn(ch1->A_dh_pub, sizeof ch1->A_dh_pub, NULL);
	if (DH_size(u->usm->dh) > sizeof g_buf)
		return -1;
	if (DH_init_keys(u->usm->kbuf, u, u->usm->dh, pub, 1) != 0)
	{
		THCerr("Failed to initialize key material\n");
		return -1;	
	}

	/*
	 * Build package that we send to Alice
	 */
	BN_bn2bin(u->usm->dh->pub_key, ch2.B_dh_pub);
	memcpy(ch2.B_noince, u->usm->B_noince, sizeof ch2.B_noince);

	b64_enc(g_buf, sizeof g_buf, (char *)&ch2, sizeof ch2);
	irc(nick, "2", "%s", g_buf);

	u->state = STATE_B_WAIT4SIGN;
	return 0;
}

/*
 * Bob send us back his public key. We can generate key material
 * from this (jappa!).
 * First time that we see Bob's noince.
 */
int
A_sign(char *nick, char *msg)
{
	USER *u;
	BIGNUM *pub;
	struct _THCch2 *ch2;
	size_t len;

	if ( (u = USER_lookup(nick)) == NULL)
	{
		THCerr("%s(): FIXME.\n", __func__);
		return -1;
	}
	if (u->state != STATE_PUBSENT)
		return -1;

	u->state = 0;

	len = b64_dec(g_buf, sizeof g_buf, msg, strlen(msg));
	if (len < sizeof *ch2)
	{
		THCerr("base64 decoding failed.\n");
		return -1;
	}
	len -= sizeof *ch2; /* signature length */

	ch2 = (struct _THCch2 *)g_buf;
	pub = BN_bin2bn(ch2->B_dh_pub, sizeof ch2->B_dh_pub, NULL);

	if (DH_size(u->usm->dh) != sizeof u->usm->kbuf)
		return -1;
	DH_init_keys(u->usm->kbuf, u, u->usm->dh, pub, 0);
	memcpy(u->usm->B_noince, ch2->B_noince, sizeof u->usm->B_noince);

	/*
	 * Send auth. packet to peer.
	 */
	snprintf(g_buf, sizeof g_buf, "%s:", opt.me);
	AB_sign(u, g_buf + opt.melen + 1, &len);

	message_enc(u, g_buf, opt.melen + 1 + len);

	u->state = STATE_A_WAIT4SIGN;
	return 0;
}

static int
AB_sign(USER *u, char *buf, size_t *len)
{
	EVP_MD_CTX md_ctx;

	EVP_SignInit(&md_ctx, EVP_md5());
	EVP_SignUpdate(&md_ctx, u->usm->A_noince, sizeof u->usm->A_noince);
	EVP_SignUpdate(&md_ctx, u->usm->B_noince, sizeof u->usm->B_noince);
	EVP_SignUpdate(&md_ctx, u->usm->kbuf, sizeof u->usm->kbuf);
	if (EVP_SignFinal(&md_ctx, buf, len, opt.privKey) != 1)
	{
		THCerr("Error creating signature\n");
		return -1;
	}

	return 0;
}

/*
 * get anick, load anick's key, verify message.
 * Return 0 on success.
 * Caller relies on the fact the anick is split wiht '\0' in the
 * same buffer.
 */
static int
AB_verify(USER *u, char *msg, char *end)
{
	EVP_MD_CTX md_ctx;
	EVP_PKEY *pkey;
	char buf[1024];
	char *ptr;
	int ret = -1;

	/*
	 * Seperate anick from <anick>:<signature> string
	 */
	ptr = msg;
	do {
		if (*ptr == ':')
			break;
	} while (ptr++ < end);
	if (ptr + 64 >= end)  /* FIXME: signature at least 64bytes! */
	{
		THCerr("Short signature\n");
		return -1;
	}
	if (ptr - msg > sizeof u->anick)
		return -1;
	*ptr++ = '\0';
	strlcpy(u->anick, msg, sizeof u->anick);

	sanitize_nick(msg, msg);
	snprintf(buf, sizeof buf, "public/thc_%s.pem", msg);
	if (!(pkey = ReadPublicKey(buf)))
	{
		THCerr("key %s: %s\n", msg, strerror(errno));
		return -1;
	}

	EVP_VerifyInit(&md_ctx, EVP_md5());
	EVP_VerifyUpdate(&md_ctx, u->usm->A_noince, sizeof u->usm->A_noince);
	EVP_VerifyUpdate(&md_ctx, u->usm->B_noince, sizeof u->usm->B_noince);
	EVP_VerifyUpdate(&md_ctx, u->usm->kbuf, sizeof u->usm->kbuf);
	if (EVP_VerifyFinal(&md_ctx, ptr, end - ptr, pkey) != 1)
	{
		THCnfo("Bad signature or wrong key from from %s (%s)\n", u->nick, msg);
		THCnfo("\"%s\" %s (%d)\n", msg, fingerprint(pkey), key_size(pkey));
		goto err;
	}

	ret = 0;
err:
	EVP_PKEY_free(pkey);

	return ret;
}

/*
 */
int
B_sign(USER *u, char *buf, size_t dlen)
{
	size_t len;

	u->state = 0;

	if (AB_verify(u, buf, buf + dlen) != 0)
		return -1;

	THCnfo("%s (%s) established a secure session.\n", u->nick, buf);
	snprintf(g_buf, sizeof g_buf, "%s:", opt.me);
	AB_sign(u, g_buf + opt.melen + 1, &len);
	message_enc(u, g_buf, opt.melen + 1 + len);

	USER_free_usm(u);
	u->state = STATE_ESTABLISHED;
	return 0;
}

/*
 * Finally we recevied the signature from Bob.
 * Discard all old data if we fail to verify him
 */
int
A_final(USER *u, char *buf, size_t dlen)
{
	u->state = 0;

	/* Get anick, verify signature */
	if (AB_verify(u, buf, buf + dlen) != 0)
		return -1;

	THCnfo("Secure session with %s (%s) established.\n", u->nick, buf);

	USER_free_usm(u);
	u->state = STATE_ESTABLISHED;
	return 0;
}


