#include "common.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/hmac.h>
#include <openssl/dh.h>
#include "thcirc.h"
#include "crypto_func.h"
#include "user.h"
#include "b64.h"

extern struct _opt opt;
static char *password;

/*
 * Read 'count' random values from a random source.
 * Returned is a pointer to buf (which must be large enough to hold
 * 'count' random bytes) or NULL on error.
 * We close the random source after each close because this function
 * is only called once per key exchange (dont waste open fd's if you
 * can waste cpu time :>).
 */
char *
get_random(char *buf, size_t count)
{
	FILE *fp;

	fp = fopen("/dev/urandom", "r");
	if (fp == NULL)
		return NULL;

	if (fread(buf, 1, count,  fp) != count)
	{
		fclose(fp);
		return NULL;
	}

	fclose(fp);
	return buf;
}

#define EVP_PKEY_XFREE(ptr)  do{if (ptr) { EVP_PKEY_free(ptr); ptr = NULL; }}while(0)


/*
 * Generate a RSA key and save them.
 */
int
gen_keys(char *pwd)
{
	RSA *rsa;
	FILE *fp;
	FILE *fp2;
	int len;
	const EVP_CIPHER *cipher = NULL;

	if (!pwd)
		return -1;

	len = strlen(pwd);
	cipher = EVP_des_ede3_cbc();

	if (opt.pubKey)
		EVP_PKEY_XFREE(opt.pubKey);

	if (opt.privKey)
		EVP_PKEY_XFREE(opt.privKey);

	rsa = RSA_generate_key(RSA_BITS,  RSA_F4, NULL, NULL);
	if (!rsa)
		return -1;
	
	fpnew(fp, "private.pem");
	fpnew(fp2, "public.pem");

	if ((!fp) || (!fp2))
	{
		if (fp)
			fclose(fp);
		if (fp2)
			fclose(fp2);
		return -1;
	}

	PEM_write_RSAPrivateKey(fp, rsa, cipher, pwd, len, NULL, NULL);
	/*
	 * stupid OpenSSL, PEM_write_RSAPublicKey() exist but does
	 * not do what we want.
	 */
	PEM_write_RSA_PUBKEY(fp2, rsa);

	fclose(fp);
	fclose(fp2);

	/* FIXME: can also extract public part of key and assign it
	 * to opt.privKey and opt.pubKey. No need to reload from
	 * file then.
	 */

	RSA_free(rsa);
	return 0;
}

/*
 * Callback function: Put the password into the array.
 * We can't pass the password directly to the private key loading
 * function as OpenSSL would then prompt the user for a password
 * if NULL is passed.
 */
int
cb_password(char *buf, int size, int rwflaw, void *pwd)
{
	if ((size <= 0) || (password == NULL))
		return 0;

	snprintf(buf, size, "%s", password);

	return strlen(password);
}

/*
 * RSA Key's are not password protected (FIXME).
 * (pass parameters to /load THCirc.irc <password here>
 * FIXME: have to check asn-routines in openssl...
 */
EVP_PKEY *
ReadPrivateKey(const char *keyfile, char *mypassword)
{
        FILE *fp = fopen(keyfile, "r");
        EVP_PKEY *pkey;

        if (!fp)
	{
		THCerr("open %s failed: %s.\n", keyfile, strerror(errno));
                return NULL;
	}

	/*
	 * pkey = (EVP_PKEY*)PEM_ASN1_read ((char *(*)())d2i_PrivateKey, PEM_STRING_EVP_PKEY, fp, NULL, NULL, NULL);
	 */
	password = mypassword;
        pkey = PEM_read_PrivateKey(fp, NULL, cb_password, NULL);

        fclose (fp);

#if 0
        if (pkey == NULL)
		ERR_print_errors_fp(stdout);
#endif

	return pkey;
}

EVP_PKEY *
ReadPublicKey(const char *certfile)
{
	FILE *fp = fopen(certfile, "r");
	EVP_PKEY *pkey;

	if (!fp)
		return NULL;

	pkey = PEM_read_PUBKEY(fp, NULL, NULL, NULL);

	fclose(fp);
	if (!pkey)
		return NULL;

	if (pkey->type != EVP_PKEY_RSA)
	{
		EVP_PKEY_free(pkey);
		return NULL;
	}

	return pkey;
}

char *
fingerprint_raw(unsigned char *data, int len)
{
	const EVP_MD *md = EVP_md5();
	EVP_MD_CTX ctx;
	unsigned char rbuf[EVP_MAX_MD_SIZE];
	int raw_len;
	static char blob[1024];
	int i;

	EVP_DigestInit(&ctx, md);
	EVP_DigestUpdate(&ctx, data, len);
	EVP_DigestFinal(&ctx, rbuf, &raw_len);

	if (sizeof blob < raw_len * 3 + 1)
		return "(key to long/broken2)";

	for (i = 0; i < raw_len; i++)
		snprintf(blob + i * 3, sizeof blob - i*3 - 3, "%02x:", rbuf[i]); 
	blob[(raw_len * 3) - 1] = '\0';

	return blob;
}

char *
fingerprint(EVP_PKEY *pkey)
{
	if (pkey->type != EVP_PKEY_RSA)
		return "(unknown key type)";

	return fingerprint_rsa(EVP_PKEY_get1_RSA(pkey));
}


char *
fingerprint_rsa(RSA *rsa)
{
	int nlen, elen;
	//static unsigned char blob[1024];
	int len;
	char blob[1024];
	char *ptr;

	nlen = BN_num_bytes(rsa->n);
	elen = BN_num_bytes(rsa->e);
	if ((nlen > 8196) || (elen > 8196))
		return "(key is broken)";

	len = nlen + elen;
	if ((len > sizeof blob) < (len < 0))
		return "(key to long/broken)";

	BN_bn2bin(rsa->n, blob);
	BN_bn2bin(rsa->e, blob + nlen);
	ptr = fingerprint_raw(blob, len);

	memset(blob, 0, len);  /* destroy memory */

	return ptr;
}

/*
 * This function always returns a positive number
 */
unsigned int
key_size(EVP_PKEY *pkey)
{
	switch (pkey->type)
	{
	case EVP_PKEY_RSA:
		return BN_num_bits(EVP_PKEY_get1_RSA(pkey)->n);
	case EVP_PKEY_DSA:
		return BN_num_bits(EVP_PKEY_get1_DSA(pkey)->p);
	}

	return 0;
}


#if 0
int
sign(char *dst, int *dlen, char *src, int slen, EVP_PKEY *pkey)
{
	EVP_MD_CTX md_ctx;

#if 0
	/*
	 * EVP_SignUpdate is doing this check for us
	 */
	if (*dlen < 64)
		return -1;
#endif

	EVP_SignInit(&md_ctx, EVP_md5());
	EVP_SignUpdate(&md_ctx, src, slen);
	if (EVP_SignFinal(&md_ctx, dst, dlen, pkey) != 1)
		return -1;

	return 0;
}

/*
 * Return 0 on success (Signature OK)
 */
int
verify(char *src, int slen, char *sig_buf, int sig_len, EVP_PKEY *pkey)
{
	EVP_MD_CTX md_ctx;

	EVP_VerifyInit(&md_ctx, EVP_md5());
	EVP_VerifyUpdate(&md_ctx, src, slen);
	if (EVP_VerifyFinal(&md_ctx, sig_buf, sig_len, pkey) != 1)
		return -1;

	return 0;
}
#endif

#if 0
void
hexdump(unsigned char *ptr, size_t len)
{
	unsigned char *end = ptr + len;

	while (ptr < end)
		printf("%2.2x ", *ptr++);
}

void hexdumpf(char *file, void *data, size_t len)
{
	FILE *fp = fopen(file, "w");

	fwrite(data, len, 1, fp);
	fclose(fp);
}
#endif

/*
 * Take binary data, build a message (encrypt + hmac) and
 * send it to the peer.
 * FIXME: src must be at least +8 bytes longer (we write data
 * into it). FIXME when implemented HMAC.
 */
int
message_enc(USER *u, char *src, size_t slen)
{
	char ebuf[2048];
	char record[2048];
	char buf[sizeof ebuf * 2];
	size_t ebuflen;
	size_t pad;
	size_t hmaclen;
	HMAC_CTX hctx;
	unsigned long seq;

	/* Who can ever send that much? */
	if (sizeof ebuf - EVP_CIPHER_block_size(ALG) < slen)
	{
		THCerr("%s(): Buffer to small.\n", __func__);
		return -1;
	}

	if (sizeof record - sizeof u->enc.hmac_seq - EVP_MAX_MD_SIZE < slen)
	{
		THCerr("%s(): MSG to long.\n", __func__);
		return -1;
	}

	/* HMAC is calculated over seq + cleartext-data and appended
	 * to the end 
	 */
	seq = htonl(u->enc.hmac_seq);
	memcpy(record, &seq, sizeof seq);
	memcpy(record + sizeof u->enc.hmac_seq, src, slen);
	src = record + sizeof u->enc.hmac_seq;

	HMAC(EVP_md5(), u->enc.hmackey, sizeof u->enc.hmackey, record, src + slen - record, src + slen, &hmaclen);
	HMAC_cleanup(&hctx);
	u->enc.hmac_seq++;
	slen += hmaclen;

	/* PKCS padding. */
	pad = EVP_CIPHER_block_size(ALG) - slen % EVP_CIPHER_block_size(ALG);
	memset(src + slen, pad, pad);
	slen += pad;

	EVP_EncryptUpdate(&u->enc.ectx, ebuf, &ebuflen, src, slen);

	b64_enc(buf, sizeof buf, ebuf, ebuflen);
	irc(u->nick, "E", "%s", buf);

	return 0;
}

/*
 * Decrypt out output
 * *dlen holds the size of dst.
 */
int
message_dec(CM *dec, char *dst, size_t *dlen, char *src, size_t slen)
{
	char buf[2048];
	char hmac[EVP_MAX_MD_SIZE];
	size_t len; 
	size_t hmaclen;
	char *ptr;
	HMAC_CTX hctx;
	unsigned long seq;

	/* IRC messages are not that big, should never happen */
	if (slen > sizeof buf)
	{
		THCerr("message to long.\n");
		return -1;
	}

	/* b64 decode */
	if ((len = b64_dec(buf, sizeof buf, src, slen)) == 0)
		return -1;

	/* decrypt */
	if ((*dlen < EVP_CIPHER_block_size(ALG)) || (*dlen - EVP_CIPHER_block_size(ALG) < len))
	{
		THCerr("Message to long.\n");
		return -1;
	}

	/* dlen is a pointer! */
	EVP_DecryptUpdate(&dec->ectx, dst, dlen, buf, len);
	ptr = dst + *dlen;
	EVP_DecryptFinal(&dec->ectx, ptr, dlen);
	*dlen += ptr - dst;

	/* verify HMAC (peer send 16 bytes of HMAC for EVP_md5() */
	if (*dlen < 16)
		return -1;

	HMAC_Init(&hctx, dec->hmackey, sizeof dec->hmackey, EVP_md5());
	seq = htonl(dec->hmac_seq);
	HMAC_Update(&hctx, (const unsigned char *)&seq, sizeof seq);
	HMAC_Update(&hctx, dst, *dlen - 16);
	HMAC_Final(&hctx, hmac, &hmaclen);
	HMAC_cleanup(&hctx);
	if (*dlen < hmaclen)
		return -1; /* Should not happen */
	if (memcmp(hmac, dst + *dlen - hmaclen, hmaclen))
	{
		THCerr("MAC wrong, Closing session.\n");
		return -1;
	}

	/* Cut away hmac (not part of message) */
	*dlen -= hmaclen;
	dst[*dlen] = '\0';

	dec->hmac_seq++;

	return 0;
}

DH *
DH_param_gen(int bits)
{
	DH *dh;
	int code;

	while (1)
	{
		dh = DH_generate_parameters(bits, DH_GENERATOR_5, NULL, NULL);
		if ((!dh) || (!DH_check(dh, &code)))
			return NULL;

		/* DH_CHECK_P_NOT_PRIME 
		 * DH_CHECK_P_NOT_SAFE_PRIME
		 * DH_UNABLE_TO_CHECK_GENERATOR
		 * DH_NOT_SUITABLE_GENERATOR
		 */
		if (code == 0)
			break;
		THCerr("%s(): First round failed, again\n", __func__);
	}

	return dh;
}

#if 0
DH *
DH_by_param(DH *dhparam)
{
	DH *dh = DH_new();

	dh->p = BN_dup(dhparam->p);
	dh->g = BN_dup(dhparam->g);

}
#endif

/*
 * The non-inverted DH shared secret is returned in kbuf.
 * The kbuf is later used to generate the signature over.
 *
 * who:
 *   alice = 0
 *   bob = 1
 *
 * Alice and Bob should not use the same key material for
 * sending and receiving (most application do). Cleartext encrypted
 * with the same key material would result in the same crypt data
 * and thus leak the information that both say the same.
 *
 * Example:
 * Alice: Hi, Bill Gates is a lamer
 * Bob  : HI, Bill Gates is really a lamer!
 *
 * The first crypt data block of both messages would be the same.
 * Eve does not know about what both parties talk about but she could
 * still say that both started the first message they exchange with the
 * same words.
 */
int
DH_init_keys(char *kbuf, USER *u, DH *dh, BIGNUM *pub, char who)
{
	char kbuf2[DH_KEY_SIZE];
	unsigned char iv[EVP_MAX_IV_LENGTH], key[EVP_MAX_KEY_LENGTH];
	int i;
	char *ptr1, *ptr2;

	if (DH_compute_key(kbuf, pub, dh) != DH_size(dh))
		return -1;
	for (i = 0; i < DH_KEY_SIZE; i++)
		kbuf2[DH_KEY_SIZE - 1 - i] = kbuf[i];

	if (who == 0)
	{
		ptr1 = kbuf;
		ptr2 = kbuf2;
	} else {
		ptr1 = kbuf2;
		ptr2 = kbuf;
	}

	EVP_BytesToKey(ALG, EVP_md5(), NULL, ptr1, DH_KEY_SIZE, 1, key, iv);
	MD5(ptr1, DH_KEY_SIZE, u->dec.hmackey);
	EVP_DecryptInit(&u->dec.ectx, ALG, key, iv);

	EVP_BytesToKey(ALG, EVP_md5(), NULL, ptr2, DH_KEY_SIZE, 1, key, iv);
	MD5(ptr2, DH_KEY_SIZE, u->enc.hmackey);
	EVP_EncryptInit(&u->enc.ectx, ALG, key, iv);

	return 0;
}


