/*
 * $Id: irc.c,v 1.2 2002/10/15 08:41:09 skyper Exp $
 */

#include "common.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include "irc.h"
#include "system.h"


static int dfrg_pos;
static struct _defrag defrag[10];

/*
 * Break a message into chunks and output them.
 * RFC says every server must accept 512 bytes of message.
 * Each message is terminated by a \r\n.
 * PRIVMSG G <NICK> :<msg>
 * len = 512 - 2 - strlen("PRIVMSG") - 5 - strlen(<NICK>);
 * HINT: The actual length is somehow magic. Various clients
 * to not support the rfc length and just truncate the buffer.
 *
 * wild tests:
 * BitchX (send): 492 chars
 * BitchX (recv): 463 chars
 * Xchat (recv) : 463 chars
 *
 * The supported length depends on the nick!nick@server string length.
 */
void
irc(char *to, char *suffix, const char *fmt, ...)
{
	char buf[1024];
	va_list ap;
	char *b, *e;
        /* we just say: no nick!user@server should be longer > 80 */
	int splitlen = 502 - 80 - sizeof("PRIVMSG") - strlen(to) - strlen(suffix);

	if (splitlen < 64)
		return;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf, fmt, ap);
	va_end(ap);

	b = buf;
	e = buf + strlen(buf);
	/* send first fragment with THCFRG_FIRST,
	 * then all following wiht THCFRG and last with THCFRG_LAST
	 */
	if (e - b > splitlen)
	{
		THCpr("%s %s %c%.*s\n", to, suffix, THCFRG_FIRST, splitlen, b);
		b+= splitlen;
		while (e - b > splitlen)
		{
			THCpr("%s %s %c%.*s\n", to, suffix, THCFRG, splitlen, b);
			b += splitlen;
		}
		THCpr("%s %s %c%s\n", to, suffix, THCFRG_LAST, b);
	} else {
		THCpr("%s %s %s\n", to, suffix, b);
	}
}

/*
 * Search backward through the ringbuffer as it's more probable
 * that the previous fragment has just been send.
 */
static struct _defrag *
dfrg_buf_lookup(char *nick)
{
	int i;

	i = dfrg_pos;
	do {
		if (strcmp(nick, defrag[i].nick) == 0)
			return &defrag[i];

		if (i == 0)
			i = sizeof defrag / sizeof *defrag;
		i--;
	} while (i != dfrg_pos);

	return &defrag[i];
}

/*
 * Defragmentation of a irc fragmented message.
 * Special care must be taken that that a stranger can not fill our USER
 * structure. Fragmented packets should be kept in a special
 * ring buffer of 50 entires which ought to be enough for everyone :>
 */
char *
defragment(char *nick, char *msg)
{
	struct _defrag *dptr;
	char *e, *ptr;

	/*
	 * Most messages do not require fragmentation.
	 * FIXME: should we then find and destroy msg
	 * in buffer? crypto attack possible?
	 */
	if (*msg == '\0')
		return NULL;

	if (*msg == THCFRG_FIRST)
	{
		dfrg_pos++;
		if (dfrg_pos >= sizeof defrag / sizeof *defrag)
			dfrg_pos = 0;

		dptr = &defrag[dfrg_pos];
		strlcpy(dptr->nick, nick, sizeof dptr->nick);
		msg++;
		dptr->len = strlcpy(dptr->buf, msg, sizeof dptr->buf);

		return NULL;
	}

	if ((*msg != THCFRG) && (*msg != THCFRG_LAST))
		return msg;

	dptr = dfrg_buf_lookup(nick);
	/* illegal! Lost THCFRG_FIRST, ring buffer to small */
	if (!dptr)
		return NULL;

	ptr = dptr->buf + dptr->len;
	e = dptr->buf + sizeof dptr->buf;
	/* Buffer to small */
	if (ptr >= e)
		return NULL;

	dptr->len = strlcat(ptr, msg + 1, e - ptr);

	if (*msg == THCFRG)
		return NULL;

	return dptr->buf;
}

