/*
 * $Id: thcirc.c,v 1.3 2002/10/15 09:24:28 skyper Exp $
 */

#include "common.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/hmac.h>
#include <openssl/dh.h>
#include <openssl/rand.h>
#include "thcirc.h"
#include "user.h"
#include "system.h"
#include "crypto_func.h"
#include "irc.h"
#include "b64.h"
#include "sanitize.h"
#include "challenge.h"

struct _opt opt;
char g_buf[4096];
char *thc_token = "testtoken";

void
init_vars(void)
{
	struct stat sbuf;

	if (stat("public", &sbuf))
	{
		if (mkpdir("public", 0700))
		{
			THCerr("mkdir \"public\": %s.\n", strerror(errno));
			exit(-1);
		}
	} else {
		if (!S_ISDIR(sbuf.st_mode))
		{
			THCerr("\"public\" is not a directory.\n");
			exit(-1);
		}
	}

	/*
	 * Should never happen unless someone fucks up the defines
	 * or changes the cipher.
	 */
	if ((EVP_CIPHER_iv_length(ALG) != IV_SIZE) || (EVP_CIPHER_key_length(ALG) != KEY_SIZE))
	{
		THCerr("%s(): INTERNAL ERROR.\n", __func__);
		exit(-1);
	}

	/*
	 * We use MD5 (16B) to generate the hmackey (16B)
	 * Otherwise use MD5 first and then use EVP_ByteToKey() to bring
	 * the MD5-output to the right size (which fits for the key).
	 */
	if ((EVP_MD_size(EVP_md5())) !=  KEY_SIZE)
	{
		THCerr("%s(): INTERNAL ERROR.\n", __func__);
		exit(-1);
	}

	if (stat("private.pem", &sbuf) != 0)
	{
		THCnfo("Use /genkey <password> to generate new keys.\n");
		return;
	}

	THCnfo("Use /loadkey <password> to load your private key.\n");
}

/*
 * Initialize the program itself (CWD, random numbers, ..).
 */
void
init_prog(void)
{
	char *ptr;

	ERR_load_crypto_strings();
	SSLeay_add_all_algorithms(); /* FIXME: dont need them all */
	RAND_seed(get_random(g_buf, sizeof g_buf), sizeof g_buf);

	ptr = getenv("THCIRC");
	if ( (!ptr) && ((ptr = getenv("HOME")) != NULL))
	{
		snprintf(g_buf, sizeof g_buf, "%s/.thcirc", ptr);
		ptr = g_buf;
	}

	if (!ptr)
	{
		THCerr("~/.thcirc does not exist.\n");
		exit(-1);
	}

	if (chdir(ptr) != 0)
	{
		THCerr("chdir(%s): %s\n", ptr, strerror(errno));
		exit(-1);
	}
}

/*
 * Return the action that should be performed
 * Return -1 if nick/msg has been specified but no value
 * can be assigned to it (missing token "!").
 * Otherwise nick/msg (if specified) are always set.
 */
int
parse(char *buf, char **nick, char **msg)
{
	char *ptr;

	ptr = strchr(buf, '!');
	if (!ptr)
		return -1;
	ptr++;

	/* Do not allow empty nick names */
	if (*ptr == '\0')
		return -1;

	*nick = ptr;

	if (!msg)
		return 0;

	ptr = strchr(ptr, '!');
	if (!ptr)
		return -1;
	*ptr++ = '\0';
	*msg = ptr;

	/* Do not allow empty messages */
	if (*msg == '\0')
		return -1;

	return 0;
}

void
do_dekey(char *nick)
{
	USER_del(nick);
}

/*
 * Output various status information
 */
void
do_info(void)
{
	USER *u = NULL;
	int n = USER_n();
	char c;
	char line[65];
	char user[65];

	THCnfo("Version: %s\n", VERSION);
	while ( (u = USER_next(u)))
	{
		if (u->flags & USER_FL_ALICE)
			c = '*';
		else
			c = ' ';

		snprintf(user, sizeof user, "%s (%s)", u->nick, u->anick);
		snprintf(line, sizeof line, "%c %3d %-40.40s %12s", c, n--, user, USER_state(u));
		THCnfo("%s\n", line);

	}
	if (opt.pubKey)
		THCnfo("\"%s\" %s (%d)\n", opt.me, fingerprint(opt.pubKey), key_size(opt.pubKey));
}

int
do_encrypt(char *nick, char *msg)
{
	USER *u;

	if ( (u = USER_lookup(nick)) == NULL)
	{
		THCnfo("Establish a secure session first. (/key %s)\n", nick);
		return -1;
	}

	if (u->state != STATE_ESTABLISHED)
	{
		THCnfo("Session not yet ready, wait.\n");
		return -1;
	}

	THCpr("%s O %s\n", nick, msg);
	message_enc(u, msg, strlen(msg));

	return 0;
}

int
do_decrypt(char *nick, char *msg)
{
	USER *u;
	size_t dlen = sizeof g_buf;
	char buf[sizeof g_buf];

	if ( (u = USER_lookup(nick)) == NULL)
	{
		THCnfo("No secure session with %s.\n", nick);
		return -1;
	}

	if (message_dec(&u->dec, buf, &dlen, msg, strlen(msg)) != 0)
		return -1;

	if (u->state == STATE_A_WAIT4SIGN)
	{
		A_final(u, buf, dlen);
		return 0;
	}

	if (u->state == STATE_B_WAIT4SIGN)
	{
		B_sign(u, buf, dlen);
		return 0;
	}

	if (u->state != STATE_ESTABLISHED)
		return -1;

	if (dlen <= 0)
		return 0;
	if (dlen >= sizeof buf)
		return 0;
	buf[dlen] = '\0';

	THCpr("%s U %s\n", nick, buf);

	return 0;
}


/*
 * Called from do_loadkey()
 */
int
load_keys(char *password)
{
	RSA *rsa;
	unsigned char *p = opt.public;
	FILE *fp;

	fp = fopen("dhparam.pem", "r");
	if (!fp)
	{
		THCerr("fopen(dhparam.pem): %s\n", strerror(errno));
		return -1;
	}
	opt.dhparam = PEM_read_DHparams(fp, NULL, NULL, NULL);
	fclose(fp);
	if (!opt.dhparam)
	{
		THCerr("Failed to load \"dhparam.pem\".\n");
		return -1;
	}

	opt.privKey = ReadPrivateKey("private.pem", password);
	memset(password, 0, strlen(password));
	if (!opt.privKey)
	{
		THCerr("Failed to load keys (wrong password?).\n");
		return -1;
	}

	opt.pubKey = ReadPublicKey("public.pem");
	if (!opt.pubKey)
	{
		THCerr("Failed to load public key.\n");
		return -1;
	}

	rsa = EVP_PKEY_get1_RSA(opt.pubKey);
	opt.publiclen = i2d_RSAPublicKey(rsa, NULL);
	if (opt.publiclen > sizeof opt.public)
	{
		EVP_PKEY_free(opt.pubKey);
		THCerr("Your public key is broken! Use /genkey.\n");
		return -1;
	}
	opt.publiclen = i2d_RSAPublicKey(rsa, &p);

	return 0;
}


int
do_loadkey(char *buf)
{
	char *pwd = NULL;
	FILE *fp;
	struct stat sbuf;

	if (parse(buf, &pwd, NULL) != 0)
		return -1;

	if (stat("private.pem", &sbuf) < 0)
	{
		THCnfo("Use /genkey <password> first.\n");
		return -1;
	}

	/*
	 * Load keys from file if password has been specified.
	 */
	if (strlen(pwd) > 0)
	{
		if (load_keys(pwd) != 0)
			return -1;

		fp = fopen("me", "r");
		if (!fp)
		{
			THCerr("fopen \"me\" failed: %s.\n", strerror(errno));
			return -1;
		}
		fgets(opt.me, sizeof opt.me, fp);
		fclose(fp);
		opt.melen = strlen(opt.me);
	}

	if (!opt.privKey)
	{
		THCnfo("Use /loadkey <password>\n");
		return -1;
	}

	THCnfo("\"%s\" %s (%d)\n", opt.me, fingerprint(opt.pubKey), key_size(opt.pubKey));

	return 0;
}

/*
 * Send our authentic nick and  public key to the remote user.
 * (authentic nick is taken from $THCIRC/me)
 */
int
do_sendkey(char *buf)
{
	char *nick;

	if ((parse(buf, &nick, NULL) != 0) || (strlen(nick) <= 0))
	{
		THCnfo("Use /sendkey <nick>\n");
		return -1;
	}

	if ((opt.privKey == NULL) || (opt.publiclen <= 0))
	{
		THCnfo("Use /loadkey <password> first.\n");
		return -1;
	}

	if (b64_enc(g_buf, sizeof g_buf, opt.public, opt.publiclen) <= 0)
		return -1;

	/*
	 * Send "anick:publickey" to nick
	 */
	irc(nick, "P", "%s:%s", opt.me, g_buf);

	return 0;
}

int
do_genkey(char *buf)
{
	char *me;
	char *pwd;
	struct stat sbuf;
	FILE *fp;
	DH *dh = NULL;

	if (parse(buf, &me, &pwd) != 0)
		return -1;

	if (strlen(pwd) <= 0)
	{
		THCnfo("Use /genkey <password>\n");
		return -1;
	}

	if (stat("private.pem", &sbuf) == 0)
	{
		THCnfo("Delete old key \"private.pem\" first.\n");
		return -1;
	}

	THCnfo("Generating %d-bit RSA keys. This may take a while...\n", RSA_BITS);
	if (gen_keys(pwd) != 0)
	{
		THCerr("Failed to generate RSA keys.\n");
		return -1;
	}

	if (stat("dhparam.pem", &sbuf) == 0)
	{
		THCnfo("Not generating new dhparam.pem. Already exist.\n");
	} else {
		THCnfo("Generating 2048-bit DH parameters. This may even take longer...\n");
		if (!(fp = fopen("dhparam.pem", "w")))
		{
			THCerr("fopen(dhparam.pem): %s\n", strerror(errno));
			return -1;
		}
		if (!(dh = DH_param_gen(2048)))
		{
			THCerr("Failed to generated DH parameters.\n");
			fclose(fp);
			return -1;
		}
		PEM_write_DHparams(fp, dh);
		fclose(fp);
		DH_free(dh);
	}

	fp = fopen("me", "w");
	if (!fp)
	{
		THCerr("fopen(\"me\"): %s\n", strerror(errno));
		return -1;
	}
	fprintf(fp, "%s", me);
	fclose(fp);

	if (load_keys(pwd) != 0)     /* Load them from harddrive again! */
		return -1;

	THCnfo("Done: \"%s\" %s (%d)\n", opt.me, fingerprint(opt.pubKey), key_size(opt.pubKey));

	return 0;
}

int
do_pubkey(char *nick, char *msg)
{
	char *ptr;
	FILE *fp;

	if ( (ptr = strchr(msg, ':')) == NULL)
		return -1;

	if ( (fp = fopen("public/unaccepted.raw", "w")) == NULL)
	{
		THCerr("fopen failed: %s\n", strerror(errno));
		return -1;
	}

	fprintf(fp, "%s:%s", nick, msg);
	*ptr = '\0';
	/*
	 * We do not show fingerprint at this point.
	 * User musst agree before we process/parse/convert any data
	 * from a untrusted user. (bad for the fingerprint :/)
	 */
	THCnfo("Use /accept to accept the public key from %s (%s).\n", nick, msg);
	fclose(fp);

	return 0;
}

/*
 * accept a key.
 */
int
do_accept(void)
{
	FILE *fp;
	int len;
	char buf[1024];
	char pkfile[1024];
	char *ptr;
	char *anick;  /* authentic nick */
	RSA *rsa;
	unsigned char *p;
	struct stat sbuf;

	if ( (fp = fopen("public/unaccepted.raw", "r")) == NULL)
		return -1;
	len = fread(g_buf, 1, sizeof g_buf, fp);
	fclose(fp);

	if (len <= 0)
		return -1;

	if ( (ptr = strchr(g_buf, ':')) == NULL)
	{
		THCerr("Public key is broken (user missing?).\n");
		return -1;
	}

	*ptr++ = '\0';
	anick = ptr;

	if ( (ptr = strchr(ptr, ':')) == NULL)
	{
		THCerr("Public key is broken.\n");
		return -1;
	}
	*ptr++ = '\0';

	len = b64_dec(buf, sizeof buf, ptr, strlen(ptr));
	if (len < 64)
	{
		THCerr("Public key is to short (%d bytes).\n", len);
		return -1;
	}

	/*
	 * Save public key from user (FIXME: check PEM_write_RSA_PUBKEY())
	 */
	sanitize_nick(anick, anick);
	snprintf(pkfile, sizeof pkfile, "public/thc_%s.pem", anick);
	/* Forbid irc junkies to send us key's for an ID we already trust */
	if (stat(pkfile, &sbuf) == 0)
	{
		THCerr("Delete old key first: \"%s\"\n", pkfile);
		return -1;
	}
	if ( (fp = fopen(pkfile, "w")) == NULL)
	{
		THCerr("fopen %s failed: %s\n", pkfile, strerror(errno));
		return -1;
	}

	if (unlink("public/unaccepted.raw"))
	{
		THCerr("unlink failed: %s.\n", strerror(errno));;
		return -1;
	}

	p = buf;
	rsa = d2i_RSAPublicKey(NULL, &p, (long)len);
	PEM_write_RSA_PUBKEY(fp, rsa);
	fclose(fp);

	THCnfo("Accpeting \"%s\" %s (%d)\n", anick, fingerprint_rsa(rsa), BN_num_bits(rsa->n));
	RSA_free(rsa);

	return 0;
}

#if 0
void
testme(void)
{
	char buf[10000];
	char dst[2 * sizeof buf];
	char final[sizeof buf + 1000];
	int ret;
	int i;

	memset(buf, 'A', sizeof buf);

	for (i = 0; i <  10000; i++)
	{
	ret = b64_enc(dst, sizeof dst, buf, i);
	DEBUGF("%d encoded data (strlen = %d)\n", ret, strlen(dst));
	if (ret != strlen(dst))
		break;
	ret = b64_dec(final, sizeof final, dst, ret);
	DEBUGF("%d decoded data\n", ret);
	if (ret != i)
		break;
	if (memcmp(buf, final, i) != 0)
		break;

	}

	exit(0);
}
#endif


#if 0
void
testdh(void)
{
	DH *dhparam;
	DH *b = DH_new();
	char *dhbuf;
	unsigned char *ptr;
	int len;
	int i;
	FILE *fp;
	DH *dh;

#if 0
	fp = fopen("dh.pem", "w");
	dh = gen_dh(1024);
	PEM_write_DHparams(fp, dh);
	fclose(fp);
	exit(0);
#endif

	if ((fp = fopen("dh.pem", "r")) == NULL)
		DEBUGF("fopen failed: %s\n", strerror(errno));
	dhparam = PEM_read_DHparams(fp, NULL, NULL, NULL);
	fclose(fp);

	if (!dhparam)
		DEBUGF("FATAL: dh gen failed\n");
	DEBUGF("DH size= %d\n", DH_size(dhparam));
	DEBUGF("p (%d)\n", BN_num_bits(dhparam->p));

	dh = DH_new();
	dh->p = BN_dup(dhparam->p);
	dh->g = BN_dup(dhparam->g);
	if (!DH_generate_key(dh))
		DEBUGF("FATAL: DH_generate_key\n");
	DEBUGF("dh->pub_key (%d) %d %s\n", BN_num_bits(dh->pub_key), BN_num_bytes(dh->pub_key), BN_bn2hex(dh->pub_key));


	dh = DH_new();
	dh->p = BN_dup(dhparam->p);
	dh->g = BN_dup(dhparam->g);
	if (!DH_generate_key(dh))
		DEBUGF("FATAL: DH_generate_key\n");
	DEBUGF("dh->pub_key (%d) %d %s\n", BN_num_bits(dh->pub_key), BN_num_bytes(dh->pub_key), BN_bn2hex(dh->pub_key));
	exit(0);

	/* transfer 32byte random noince */
	/* transfer p and dh->pub_key to peer. (g is always 5) */
/* PEER START */
	if (!b)
		DEBUGF("FATAL: DH_new()\n");

	b->p = BN_dup(dh->p);
	b->g = BN_new();
	BN_hex2bn(&b->g, "05");
	if (!DH_generate_key(b))
		DEBUGF("FATAL: DH_generate_key(b)\n");

	ptr = malloc(DH_size(b));
	if (DH_compute_key(ptr, dh->pub_key, b) != DH_size(b))
		DEBUGF("FATAL: DH_compute_key\n");

	fprintf(stderr, "bkey1  = ");
	for (i = 0; i < DH_size(b); i++)
		fprintf(stderr, "%02X", ptr[i]);
	fprintf(stderr, "\n");
	DEBUGF("b->pub_key (%d)\n", BN_num_bits(b->pub_key));
/* PEER END */
	/* transfer b->pub_key to the peer */

	if (DH_compute_key(ptr, b->pub_key, dh) != DH_size(dh))
		DEBUGF("FATAL: DH_compute_key(dh)\n");

	fprintf(stderr, "dhkey1 = ");
	for (i = 0; i < DH_size(dh); i++)
		fprintf(stderr, "%02X", ptr[i]);
	fprintf(stderr, "\n");

err:
	ERR_print_errors_fp(stderr);
}
#endif

void
testsig(void)
{
	char data[1024];
	EVP_PKEY *pkey = ReadPublicKey("public/thc_skyper.pem");
	EVP_MD_CTX md_ctx;
	int len;
	char buf[1024];

	opt.privKey = ReadPrivateKey("private.pem", "lala");
	memset(data, 0x41, sizeof data);
	EVP_SignInit(&md_ctx, EVP_md5());
	EVP_SignUpdate(&md_ctx, data, sizeof data);
	DEBUGF("hi %p\n", opt.privKey);
	EVP_SignFinal(&md_ctx, buf, &len, opt.privKey);
	DEBUGF("hi\n");
	printf("%d signdata\n", len); /* TEST */

	EVP_VerifyInit(&md_ctx, EVP_md5());
	EVP_VerifyUpdate(&md_ctx, data, sizeof data);
	len = EVP_VerifyFinal(&md_ctx, buf, len, pkey);
	printf("Verify returned %d\n", len); /* TEST */

}


int
main(int argc, char *argv[])
{
	char *buf;
	char line[1024];
	char *ptr;
	char *nick, *msg;

	if (argc != 1)
	{
		printf("This is an IRC-II crypto addon. Use /load THCirc.irc.\n");
		exit(0);
	}

	init_prog();
	init_vars();
	//setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stdout, NULL, _IOFBF, 0); /* Fully buffered */
#if 0
	testsig();
	exit(0);
#endif
#if 0
	testdh();
	exit(0);
#endif

	thc_token = line;

	while (fgets(line, sizeof line, stdin))
	{
		if ( (ptr = strchr(line, '!')) == NULL)
			continue;

		*ptr++ = '\0';
		buf = ptr;

		if ( (ptr = strchr(buf, '\n')) == NULL)
			continue;
		*ptr = '\0';


		switch (buf[0])
		{
		case 'N':
			do_info();
			break;
		case 'A':
			do_accept();
			break;
		case 'G':
			do_genkey(buf);
			break;
		case 'L':
			do_loadkey(buf);
			break;
		case 'S':
			do_sendkey(buf);
			break;
		case 'D':
			if (parse(buf, &nick, NULL) != 0)
				break;
			do_dekey(nick);
			break;
		case 'P':
			if (parse(buf, &nick, &msg) != 0)
				break;
			sanitize_nick(nick, nick);
			if ((msg = defragment(nick, msg)) == NULL)
				break;
			do_pubkey(nick, msg);
			break;

		/* Handshake */
		case 'C':
			if (parse(buf, &nick, NULL) != 0)
				break;
			A_send(nick);
			break;
		case 'K':  /* PEER: Alice's DH parameters */
			if (parse(buf, &nick, &msg) != 0)
				break;
			sanitize_nick(nick, nick);
			if ((msg = defragment(nick, msg)) == NULL)
				break;
			B_answer(nick, msg);
			break;
		case '2': /* US: received the challenge answer, signed */
			if (parse(buf, &nick, &msg) != 0)
				break;
			sanitize_nick(nick, nick);
			if ((msg = defragment(nick, msg)) == NULL)
				break;
			A_sign(nick, msg);
			break;
		/* Secure Session */
		case 'E': /* encrypt privmsg and send to peer */
			if (parse(buf, &nick, &msg) != 0)
				break;
			do_encrypt(nick, msg);
			break;
		case 'U': /* receive crypted msg and print out */
			if (parse(buf, &nick, &msg) != 0)
				break;
			sanitize_nick(nick, nick);
			if ((msg = defragment(nick, msg)) == NULL)
				break;
			do_decrypt(nick, msg);
			break;
		default:
			THCerr("THCirc: Unknown Command.\n");
		}
	}

	exit(0);
	return 0;
}

